package org.processmining.lpm.discovery;

import static org.processmining.lpm.util.AcceptingPetrinetUtils.decodeTransitionLabels;
import static org.processmining.lpm.util.LogUtils.encodeLogUsingScheme;
import static org.processmining.lpm.util.LogUtils.getActivityCounts;
import static org.processmining.lpm.util.LogUtils.getLogDecodingScheme;
import static org.processmining.lpm.util.LogUtils.getLogEncodingScheme;
import static org.processmining.lpm.util.ProcessTreeUtils.decodeTransitionLabels;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ForkJoinPool;

import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.lpm.dialogs.LocalProcessModelParameters;
import org.processmining.lpm.dialogs.LocalProcessModelParameters.ProjectionMethods;
import org.processmining.lpm.util.LPMXLog;
import org.processmining.lpm.util.LocalProcessModel;
import org.processmining.lpm.util.LocalProcessModelEvaluator;
import org.processmining.lpm.util.LocalProcessModelRanking;
import org.processmining.lpm.util.LocalProcessModelRankingFactory;
import org.processmining.lpm.util.LocalProcessModelTopSet;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.impl.ProcessTreeImpl;

@Plugin(
		name = "Search for Local Process Models", 
		parameterLabels = {"Input Log"}, 
	    returnLabels = {"Local Process Model Ranking"}, 
	    returnTypes = { LPMXLog.class },
	    level = PluginLevel.PeerReviewed
		)
public class LocalProcessModelDiscovery {
	private LocalProcessModelTopSet topSet;

	private Map<String, Character> encodingScheme;
	private Map<Character, String> decodingScheme;
	private Set<Character> transitions;
	
	private Set<String> lpmTreeSet;
	
	private Map<String, Integer> logActivityCountsMap;
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl")
	@PluginVariant(variantLabel = "Search for Local Process Models", requiredParameterLabels = {0})
	public LPMXLog run(UIPluginContext context, XLog log) {
		//params = new LocalProcessModelParameters();
		//params.setDiscoveryLog(log);
		//params.setEvaluationLog(log);
		//params.setSmartParameterDefaultsForDiscoveryLog();
		//LocalProcessModelDialog dialog = new LocalProcessModelDialog(params);
	    // Show the dialog. User can now change the configuration.
	    //InteractionResult result = context.showWizard("Configure parameters for Local Process Model Discovery", true, true, dialog);
	    // User has close the dialog.
	    //if (result == InteractionResult.FINISHED) 
	    //	return runHeadless(context, params);
		return new LPMXLog(log);
	}
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl")
	@PluginVariant(variantLabel = "Search for Local Process Models", requiredParameterLabels = {0})
	public LPMXLog run(PluginContext context, XLog log) {
		//params = new LocalProcessModelParameters();
		//params.setDiscoveryLog(log);
		//params.setEvaluationLog(log);
		//params.setSmartParameterDefaultsForDiscoveryLog();
		//return runHeadless(context, params); // uses default parameters
		//LocalProcessModelMiner miner = new LocalProcessModelMiner();
		return new LPMXLog(log);
	}
	
	public LocalProcessModelRanking runHeadlessProjectionless(PluginContext context, LocalProcessModelParameters params){
		encodingScheme = getLogEncodingScheme(params.getDiscoveryLog());
		decodingScheme = getLogDecodingScheme(encodingScheme);
		transitions = decodingScheme.keySet();
		
		XLog evaluationLog = params.getEvaluationLog();
		params.setDiscoveryLog(encodeLogUsingScheme(params.getDiscoveryLog(), encodingScheme, params.getAttributesToKeep()));
		params.setEvaluationLog(params.getDiscoveryLog());
		this.logActivityCountsMap = getActivityCounts(params.getDiscoveryLog());
		
		topSet = new LocalProcessModelTopSet(params.getTop_k());
		
		ProcessTree pt = new ProcessTreeImpl();
		
		lpmTreeSet = Collections.synchronizedSet(new HashSet<String>());
		LocalProcessModelEvaluator scorer = new LocalProcessModelEvaluator(params);

		ForkJoinPool p = new ForkJoinPool(Runtime.getRuntime().availableProcessors()-1);
		LPMRecursiveAction root = new LPMRecursiveAction(context, params, lpmTreeSet, pt, 0, null, transitions, topSet, logActivityCountsMap, scorer);
		
		p.invoke(root);
		
		params.setEvaluationLog(evaluationLog);
		scorer.refreshEfficientLogComponents();
		
		LocalProcessModelTopSet resultTopSet = new LocalProcessModelTopSet(topSet.getMaxSize());
		if(params.isVerbose())
			System.out.println("Final selection: ");
		Map<String, Integer> logActivityCountsMapNonEncoded = getActivityCounts(evaluationLog);
		for(LocalProcessModel lpm : topSet){
			// decode Petri nets
			AcceptingPetriNet net2 = decodeTransitionLabels(lpm.getAcceptingPetriNet(), decodingScheme);
			LocalProcessModel murataLpm = scorer.evaluateNetOnLog(context, net2, logActivityCountsMapNonEncoded, false);
			if(params.isStoreProcessTree()){
				murataLpm.setProcessTree(decodeTransitionLabels(lpm.getProcessTree(), decodingScheme));
			}
			murataLpm.setLogActivityCountMap(logActivityCountsMapNonEncoded);
			resultTopSet.add(murataLpm); // needed to resort the Murata versions of the nets (determinism score might be slightly different)
		}
		
		LocalProcessModelRanking finalResult = LocalProcessModelRankingFactory.createCountedAcceptingPetriNetArray();
		for(LocalProcessModel lpm : resultTopSet)
			finalResult.addElement(lpm);
		if(params.isVerbose())
			System.out.println(finalResult.getSize()+" Petri nets selected out of "+params.getNumberOfExploredLpms().get()+" candidates");
		params.setDiscoveryLog(evaluationLog);
		return finalResult;
	}
	
	public LocalProcessModelRanking runHeadless(PluginContext context, LocalProcessModelParameters params){
    	if(params.getProjectionMethod().equals(ProjectionMethods.None))
    		return runHeadlessProjectionless(context, params);
    	else if(params.getProjectionMethod().equals(ProjectionMethods.Markov))
    		return MarkovBasedSearch.runHeadless(context, 0, params, this, 1.5);
    	else if(params.getProjectionMethod().equals(ProjectionMethods.MRIG))
    		return MRIGBasedSearch.runHeadless(context, params, this);
    	else if(params.getProjectionMethod().equals(ProjectionMethods.Entropy))
    		return EntropyBasedSearch.runHeadless(context, params, this);
    	return null;
	}
}